# -*- coding: UTF-8 -*-
# Copyright 2020 by Rumma & Ko Ltd.
# License: BSD, see file LICENSE for more details.

"""
This defines the :manage:`passwd` management command.

See :doc:`/dev/users`.

"""

import djclick as click
import getpass

from lino.api import rt

def default_username():
    return getpass.getuser()

# copied from getlino.utils:
BATCH_HELP = "Whether to run in batch mode, i.e. without asking any questions.  "\
             "Don't use this on a machine that is already being used."

# copied from getlino.utils:
def yes_or_no(msg, yes="yY", no="nN", default=True):
    """Ask for confirmation without accepting a mere RETURN."""
    click.echo(msg + " [y or n]", nl=False)
    while True:
        c = click.getchar()
        if c in yes:
            click.echo(" Yes")
            return True
        elif c in no:
            click.echo(" No")
            return False


@click.command()
@click.pass_context
@click.argument('username', default=default_username)
@click.option('-c', '--create/--no-create', default=False, help="""
Create the given user. Fail if that username eixsts already.
""")
@click.option('--batch/--no-batch', default=False, help=BATCH_HELP)
def command(ctx, username, create, batch):
    """
    Update or optionally create password, name and type of a user.

    USERNAME : The username of the user to process.

    """
    User = rt.models.users.User
    UserTypes = rt.models.users.UserTypes
    try:
        user = User.objects.get(username=username)
        if create:
            raise click.UsageError("Cannot create existing user named '{}'".format(username))
    except User.DoesNotExist:
        if create:
            user = User(username=username)
            click.echo("Creating new user")
        else:
            raise click.UsageError("We have no user named '{}'".format(username))

    dut = user.user_type or UserTypes.user
    if not batch:
        user.first_name  = click.prompt("First name", default=user.first_name or username)
        user.last_name  = click.prompt("Last name", default=user.last_name or username)
        user_type  = click.prompt("User type", type=click.Choice(
            [ut.name for ut in UserTypes.get_list_items() if ut.name]),
            default=dut.name)
        if user_type:
            user.user_type = UserTypes.get_by_name(user_type)
        else:
            user.user_type = None
        passwd = click.prompt(
            "Password (leave blank to deactivate user account)",
            hide_input=True, confirmation_prompt=True, default='')
        if passwd:
            user.set_password(passwd)
        else:
            user.set_unusable_password()
    user.full_clean()
    if batch or yes_or_no('Going to save {}. Are you sure?'.format(user)):
        user.save()
        click.echo("User {} has been saved.".format(user))



# -*- coding: UTF-8 -*-
# Copyright 2014-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from builtins import map
from builtins import object

from django.conf import settings
from django.db import models
from django.contrib.contenttypes.fields import GenericRelation

from lino.api import dd, rt, _
from lino.core.actions import SubmitInsert
from lino.core.gfks import gfk2lookup
from lino.core.gfks import ContentType


class CheckedSubmitInsert(SubmitInsert):
    """Like the standard :class:`lino.core.actions.SubmitInsert`, but adds
    a confirmation if there is a possible duplicate record.

    """
    def run_from_ui(self, ar, **kw):
        obj = ar.create_instance_from_request()

        def ok(ar2):
            self.save_new_instance(ar2, obj)
            ar2.set_response(close_window=True)
            # logger.info("20140512 CheckedSubmitInsert")

        qs = list(obj.find_similar_instances(4))
        if len(qs) > 0:
            msg = _("There are %d similar %s:") % (
                len(qs), obj._meta.verbose_name_plural)
            for other in qs:
                msg += '<br/>\n' + str(other)

            msg += '<br/>\n'
            msg += _("Are you sure you want to create a new "
                     "%(model)s named %(name)s?") % dict(
                model=obj._meta.verbose_name,
                name=obj.get_full_name())

            ar.confirm(ok, msg)
        else:
            ok(ar)

class Dupable(dd.Model):
    """Base class for models that can be "dupable".

    This mixin is to be used on models for which there is a danger of
    having unwanted duplicate records. It is both for *avoiding* such
    duplicates on new records and for *detecting* existing duplicates.

    """
    class Meta(object):
        abstract = True

    dupable_words = GenericRelation(
        'dupable.PhoneticWord',
        content_type_field='owner_type',
        object_id_field='owner_id')

    # show_words = dd.ShowSlaveTable('dupable.WordsByOwner')

    submit_insert = CheckedSubmitInsert()
    """A dupable model has its
    :attr:`submit_insert<lino.core.model.Model.submit_insert>` action
    overridden by :class:`CheckedSubmitInsert`, a extended variant of
    the action which checks for duplicate rows and asks a user
    confirmation when necessary.

    """

    dupable_words_field = 'name'
    """The name of a CharField on this model which holds the full-text
    description that is being tested for duplicates."""

    def dupable_matches_required(self):
        """Return the minimum number of words that must sound alike before
        two rows should be considered similar.

        """
        return 2

    def update_dupable_words(self, really=True):
        """Update the phonetic words of this row."""

        # Excerpt from Django docs: "A related object set can be
        # replaced in bulk with one operation by assigning a new
        # iterable of objects to it". But only when the relation is
        # nullable...
        if settings.SITE.loading_from_dump:
            return
        PhoneticWord = rt.models.dupable.PhoneticWord
        qs = PhoneticWord.objects.filter(
            **gfk2lookup(PhoneticWord.owner, self)).order_by('id')
        existing = [o.word for o in qs]
        wanted = list(self.get_dupable_words(
            getattr(self, self.dupable_words_field)))
        if existing == wanted:
            return
        if really:
            qs.delete()
            for w in wanted:
                PhoneticWord(word=w, owner=self).save()
        return _("Must update phonetic words.")
        # return _("(existing {}, wanted {})").format(existing, wanted)

    def after_ui_save(self, ar, cw):
        super(Dupable, self).after_ui_save(ar, cw)
        if cw is None or cw.has_changed(self.dupable_words_field):
            self.update_dupable_words()

    def get_dupable_words(self, s):
        for c in '-,/&+':
            s = s.replace(c, ' ')
        return filter(
            None, map(
                rt.models.dupable.PhoneticWord.reduce_word,
                s.split()))

    def find_similar_instances(self, limit=None, **kwargs):
        """Return a queryset or yield a list of similar objects.

        If `limit` is specified, we never want to see more than
        `limit` duplicates.

        Note that an overridden version of this method might return a
        list or generator instead of a Django queryset.

        """
        qs = self.__class__.objects.filter(**kwargs)
        if self.pk is not None:
            qs = qs.exclude(pk=self.pk)
        parts = self.get_dupable_words(
            getattr(self, self.dupable_words_field))
        if False:
            fkw = gfk2lookup(rt.models.dupable.PhoneticWord.owner, self)
            wq = rt.models.dupable.PhoneticWord.objects.filter(**fkw)
            wq = wq.filter(word__in=parts).distinct()
            qs = qs.annotate(num=models.Count('dupable_words__word'))
            # ct = ContentType.objects.get_for_model(self.__class__)
            # qs = qs.filter(owner_type=ct)
        if True:
            qs = qs.filter(dupable_words__word__in=parts).distinct()
            qs = qs.annotate(num=models.Count('dupable_words__word'))
        qs = qs.filter(num__gte=self.dupable_matches_required())
        qs = qs.order_by('-num', 'pk')
        # print("20150306 find_similar_instances %s" % qs.query)
        if limit is None:
            return qs
        return qs[:limit]





from lino.modlib.checkdata.choicelists import Checker


class DupableChecker(Checker):
    """Checks for the following repairable problem:

    - :message:`Must update phonetic words.`

    """
    verbose_name = _("Check for missing phonetic words")
    model = Dupable

    def get_checkdata_problems(self, obj, fix=False):
        msg = obj.update_dupable_words(fix)
        if msg:
            yield (True, msg)


DupableChecker.activate()

class SimilarObjectsChecker(Checker):
    model = Dupable
    verbose_name = _("Check for similar objects")

    def get_checkdata_problems(self, obj, fix=False):
        lst = list(obj.find_similar_instances(1))
        if len(lst):
            msg = _("Similar clients: {clients}").format(
                clients=', '.join(map(str, lst)))
            yield (False, msg)

SimilarObjectsChecker.activate()



# -*- coding: UTF-8 -*-
# Copyright 2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from lino.api import dd, rt, _

from django import http
from django.conf import settings
from django.utils.translation import get_language
from lino.modlib.printing.mixins import Printable
from lino.modlib.printing.choicelists import BuildMethods

from inspect import isclass

class PreviewPublication(dd.Action):
    label = _("Preview")
    select_rows = False

    def run_from_ui(self, ar, **kw):
        sr_selected = not isclass(self)
        if sr_selected:
            ar.success(open_url=self.publisher_url())
        else:
            ar.success(open_url=self.publisher_url(self, not sr_selected))

    def get_view_permission(self, user_type):
        if not dd.is_installed('publisher'):
            return False
        return super(PreviewPublication, self).get_view_permission(user_type)


class Publishable(Printable):
    class Meta:
        abstract = True
        app_label = 'publisher'

    publisher_location = None
    publisher_template = "publisher/default.pub.html"
    publisher_list_template = "publisher/default_list.pub.html"

    list_template = "publisher/default_list_item.html"

    preview_publication = PreviewPublication()

    # @dd.action(select_rows=False)
    # def preview_publication(self, ar):
    #     sr_selected = not isclass(self)
    #     if sr_selected:
    #         ar.success(open_url=self.publisher_url())
    #     else:
    #         ar.success(open_url=self.publisher_url(self, not sr_selected))

    def publisher_url(self, list=False):
        if list:
            return "/{}/".format(self.publisher_location)
        return "/{}/{}".format(self.publisher_location, self.pk)

    @classmethod
    def get_publisher_response(cls, ar, obj=None):
        env = settings.SITE.plugins.jinja.renderer.jinja_env
        template = env.get_template(cls.publisher_template if obj else cls.publisher_list_template)
        context = ar.get_printable_context(obj=obj, model=cls)
        # context = dict(obj=self, request=request, language=get_language())
        response = http.HttpResponse(
            template.render(**context),
            content_type='text/html;charset="utf-8"')
        return response

    @classmethod
    def get_dashboard_objects(cls, user):
        return []



# Copyright 2009-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

"""Defines the :class:`Instantiator` class and some other utilities
used for generating database objects in :ref:`python fixtures <dpy>`.

"""

# from builtins import object

# import logging ; logger = logging.getLogger(__name__)

from lino.core.utils import resolve_model, UnresolvedModel
from lino.core.utils import obj2str
from lino.utils import i2d  # for backward compatibility of .py fixtures
from lino.core.fields import make_remote_field, RemoteField
from lino.utils.choosers import make_converter


class Instantiator(object):
    """A utility class to make python fixtures more compact. See
    :ref:`tutorial.instantiator`.

    An instantiator is a

    """

    def __init__(self, model, fieldnames=None, converter_classes={}, **kw):
        # self.model = resolve_model(model,strict=True)
        self.model = resolve_model(model)
        if isinstance(self.model, UnresolvedModel):
            raise Exception("Instantiator on unresolved model %s", model)
        if not self.is_active():
            def noop(*values, **kw):
                pass
            self.build = noop
            return
        # if self.model._meta.pk is None:
            # raise Exception("Model %r is not installed (_meta.pk is None)." % self.model)
        # if type(fieldnames) == str:
        if isinstance(fieldnames, str):
            fieldnames = fieldnames.split()
        self.default_values = kw
        #self.fieldnames = fieldnames
        lookup_fields = {}
        self.converters = []
        if fieldnames is None:
            self.fields = self.model._meta.fields
        else:
            self.fields = []
            for name in fieldnames:
                a = name.split(":")
                if len(a) == 2:
                    name = a[0]
                    lookup_fields[name] = a[1]

                rf = make_remote_field(model, name)
                if rf:
                    self.fields.append(rf)
                else:
                    field = self.model._meta.get_field(name)
                    self.fields.append(field)
        # print " ".join(dir(model_class))
        # print " ".join(model_class._meta.fields)
        # for f in model_class._meta.fields:
        # for f in self.fields:
        for f in self.model._meta.fields + self.model._meta.many_to_many:
            cv = None
            cvc = converter_classes.get(f.name, None)
            if cvc is not None:
                cv = cvc(f)
            else:
                cv = make_converter(f, lookup_fields)
            if cv is not None:
                self.converters.append(cv)
        # for f in model_class._meta.many_to_many:
            # print "foo", f.name

    def is_active(self):

        if isinstance(self.model, UnresolvedModel):
            return False
        if self.model._meta.pk is None:
            return False
        return True

    def __call__(self, *values, **kw):
        """Calling an instantiator is the same as callig its :meth:`build`
        method.

        """
        return self.build(*values, **kw)

    def build(self, *values, **kw):
        """Instantiate an object using the default values of this
        instantiator, overridden by the given specified values. The
        number of positional arguments may not exceed the number of
        fieldnames specified when creating this :class:`Instantiator`.

        """
        # logger.debug("Instantiator.build(%s,%r,%r)",self.model_class._meta.db_table,values,kw)
        # i = 0
        kw['_m2m'] = {}
        setters = []
        for i, v in enumerate(values):
            fld = self.fields[i]
            if isinstance(v, str):
                v = v.strip()
                if len(v) == 0:
                    continue
            if isinstance(fld, RemoteField):
                setters.append((fld, v))
            else:
                kw[fld.name] = v

        for name, v in kw.items():
            if "__" in name:
                del kw[name]
                fld = make_remote_field(self.model, name)
                setters.append((fld, v))

        # kw.update(self.default_values)
        for k, v in self.default_values.items():
            kw.setdefault(k, v)
        for c in self.converters:
            kw = c.convert(**kw)
        # if self.model.__name__ == 'Company':
            # print 20130212, __file__, kw
            # logger.info("20130212 field_cache for %s (%s)",self.model,
              # ' '.join([f.name for f in self.model._meta._field_name_cache]))

        m2m = kw.pop("_m2m")
        # print("20191210 {}({})", self.model, kw)
        instance = self.model(**kw)
        instance.full_clean()
        if m2m or len(setters):
            instance.save()

        for fld, v in setters:
            fld.setter(instance, v)

        if m2m:
            for k, v in m2m.items():
                queryset = getattr(instance, k)
                queryset.add(*v)
        return instance


class InstanceGenerator(object):
    """
    Usage example see :mod:`lino_xl.lib.humanlinks.fixtures`.
    """
    def __init__(self):
        self._objects = []
        self._instantiators = dict()

    def add_instantiator(self, name, *args, **kw):
        i = Instantiator(*args, **kw)
        # self._instantiators[i.model] = i

        def f(*args, **kw):
            o = i.build(*args, **kw)
            return self.on_new(o)
        setattr(self, name, f)

    def on_new(self, o):
        self._objects.append(o)
        return o

    def flush(self):
        rv = self._objects
        self._objects = []
        return rv


def create_row(model, **kw):
    """Instantiate, full_clean, save and return a database object.

    """
    # model = resolve_model(model)
    o = model(**kw)
    o.full_clean()
    o.save()
    return o

def create_or_update_row(model, lookup_values, new_values):
    """
    Update a single existing row that can be found using param: lookup values, with new_values.
    If one doesn't exist, instantiates, full_cleans, saves and returns a database object
    """
    try:
        o = model.objects.get(**lookup_values)
        for attr, value in new_values.items():
            setattr(o, attr, value)
    except model.DoesNotExist: # If get receives more than one item will raise MultipleObjectsReturned:
        o = model(**new_values)
    o.full_clean()
    o.save()
    return o


create = create_row  # backwards-compat

def create_and_get(model, **kw):
    """Create and then read back from database (the latter to avoid
    certain Django side effects)

    """
    o = create(model, **kw)
    return model.objects.get(pk=o.pk)

def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()



# -*- coding: UTF-8 -*-
# Copyright 2010-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)
"""
This defines the :class:`Registable` model mixin.
"""

from django.db import models
from django.utils.translation import ugettext_lazy as _

from lino.core import model
from lino.core.actions import Action
from lino.core.workflows import ChangeStateAction
from lino.core.exceptions import ChangedAPI

from lino.core.workflows import State



class RegistrableState(State):
    """

    Base class for the choices of the choicelist that defines the :attr:`state`
    field of any :class:`Registrable`.

    """
    is_editable = True
    """
    Whether the registrable object having this state should be
    editable or not.
    """



class Registrable(model.Model):

    """

    Base class to anything that may be "registered" and "deregistered", where
    "registered" means "this object has been taken account of".

    For example, when a :term:`ledger voucher` is registered, its associated
    :term:`ledger movements <ledger movement>` have been generated.
    Deregistering a voucher will first delete these movements.

    Registered objects are usually not editable, but this readonlyness is
    usually limited to certain fields.

    :class:`lino_xl.lib.cal.Reservation` is an example of a registrable that is
    not a ledger voucher.

    Subclasses must themselves define a field :attr:`state`.

    .. attribute:: state

        The workflow state field.

        This field must be a choicelist field, and the choices representing
        these states must be subclasses of :class:`RegistrableState`.

        There must be one state named "draft", which will be set e.g. after
        duplicating a registered object.

        There is no need to have a state named "registered". Actually
        "registered" means "in a non editable state".

    """
    class Meta(object):
        abstract = True

    workflow_state_field = 'state'

    _registrable_fields = None

    @classmethod
    def on_analyze(cls, site):
        super(Registrable, cls).on_analyze(site)
        if cls.workflow_state_field is None:
            raise Exception("{} has no workflow_state_field".format(cls))
        chl = cls.workflow_state_field.choicelist
        # for k in ('draft', 'registered'):
        #     if not hasattr(chl, k):
        #         raise Exception("{} has no value '{}'".format(chl, k))
        ic = chl.item_class
        if not issubclass(chl.item_class, RegistrableState):
            raise Exception("Invalid choicelist {} for {} state".format(chl, cls))
        k = 'is_editable'
        if not hasattr(ic, k):
            fld = getattr(chl, k, None)
            if not isinstance(fld, models.BooleanField):
                raise ChangedAPI(
                    "The workflow state field for {} uses {} which "
                    "has no attribute {}".format(cls, ic, k))
        cls._registrable_fields = set(cls.get_registrable_fields(site))

    @classmethod
    def get_registrable_fields(cls, site):
        """
        Return a list of the fields that are *disabled* when this is
        *registered* (i.e. `state` is not `editable`).

        Usage example::

            class MyModel(dd.Registrable):

                @classmethod
                def get_registrable_fields(self, site):
                    for f in super(MyModel, self).get_registrable_fields(site):
                        yield f
                    yield 'user'
                    yield 'date'
        """
        return []
        # yield 'date'

    def disabled_fields(self, ar):
        if not self.state.is_editable:
            return self._registrable_fields
        return super(Registrable, self).disabled_fields(ar)

    def get_row_permission(self, ar, state, ba):
        """Only rows in an editable state may be edited.

        Note that `ba` is the action being requested while
        `ar.bound_action` is the action from which the request was
        started.

        """
        if state and not state.is_editable and not isinstance(
                ba.action, ChangeStateAction):
            # if not ar.bound_action.action.readonly:
            if not ba.action.readonly:
                return False
        return super(Registrable, self).get_row_permission(ar, state, ba)

    def register(self, ar):
        """
        Register this object.  The base implementation just sets the state
        to "registered".

        Subclasses may override this to add custom behaviour.  Instead
        of subclassing you can also override :meth:`set_workflow_state
        <lino.core.model.Model.set_workflow_state>`,
        :meth:`before_state_change
        <lino.core.model.Model.before_state_change>` or
        :meth:`after_state_change
        <lino.core.model.Model.after_state_change>`.
        """

        # state_field = self._meta.get_field(self.workflow_state_field)
        state_field = self.workflow_state_field
        target_state = state_field.choicelist.registered
        self.set_workflow_state(ar, state_field, target_state)

    def deregister(self, ar):
        """
        Deregister this object.  The base implementation just sets the
        state to "draft".

        Subclasses may override this to add custom behaviour.  Instead
        of subclassing you can also override :meth:`set_workflow_state
        <lino.core.model.Model.set_workflow_state>`,
        :meth:`before_state_change
        <lino.core.model.Model.before_state_change>` or
        :meth:`after_state_change
        <lino.core.model.Model.after_state_change>`.
        """

        # state_field = self._meta.get_field(self.workflow_state_field)
        state_field = self.workflow_state_field
        target_state = state_field.choicelist.draft
        self.set_workflow_state(ar, state_field, target_state)

    # no longer needed after 20170826
    # @classmethod
    # def setup_parameters(cls, **fields):
    #     wsf = cls.workflow_state_field
    #     fields[wsf.name] = wsf.choicelist.field(blank=True, null=True)
    #     return super(Registrable, cls).setup_parameters(**fields)

    @classmethod
    def get_simple_parameters(cls):
        for p in super(Registrable, cls).get_simple_parameters():
            yield p
        # if isinstance(cls.workflow_state_field, str):
        #     raise Exception("Unresolved workflow state field in {}".format(cls))
        yield "state"
        # yield cls.workflow_state_field.name fails e.g. for vat.VatInvoices
        # because the actor is on an abstract model


    def on_duplicate(self, ar, master):
        self.state = self.workflow_state_field.choicelist.draft
        super(Registrable, self).on_duplicate(ar, master)



# -*- coding: UTF-8 -*-
# Copyright 2002-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

"""Some extensions for Sphinx.

.. autosummary::
   :toctree:

   base
   logo
   actordoc
   help_texts_extractor

"""

from atelier import sphinxconf

import sys
from sphinx.ext import autosummary
from typing import Any, Dict, List, Tuple
from sphinx.ext.autodoc.importer import import_module
from sphinx.util import logging ; logger = logging.getLogger(__name__)


# patch for autosummary. This version of import_by_name doesn't swallow the traceback
# and imports only the first item prefixes
# def my_import_by_name(name: str, prefixes: List[str] = [None]) -> Tuple[str, Any, Any, str]:
#     """Import a Python object that has the given *name*, under one of the
#     *prefixes*.  The first name that succeeds is used.
#     """
#     for prefix in prefixes:
#         if prefix:
#             prefixed_name = '.'.join([prefix, name])
#         else:
#             prefixed_name = name
#         obj, parent, modname = _import_by_name(prefixed_name)
#         return prefixed_name, obj, parent, modname

# temporary patch reports failed imports
def _import_by_name(name: str) -> Tuple[Any, Any, str]:
    """Import a Python object given its full name."""

    try:
        name_parts = name.split('.')

        # # 20200429 first try whether it is a full module name
        # try:
        #     mod = import_module(name)
        #     parent_name = '.'.join(name_parts[:-1])
        #     if parent_name:
        #         parent = import_module(parent_name)
        #     else:
        #         parent = None
        #     return parent, mod, name
        # except ImportError:
        #     pass

        # try first interpret `name` as MODNAME.OBJ
        modname = '.'.join(name_parts[:-1])
        if modname:
            try:
                mod = import_module(modname)
                return getattr(mod, name_parts[-1]), mod, modname
            except (ImportError, IndexError, AttributeError):
                pass

        # ... then as MODNAME, MODNAME.OBJ1, MODNAME.OBJ1.OBJ2, ...
        last_j = 0
        modname = None
        for j in reversed(range(1, len(name_parts) + 1)):
            last_j = j
            modname = '.'.join(name_parts[:j])
            try:
                import_module(modname)
            except ImportError as e:
                logger.info("Failed to import %s : %s", modname, e)
                continue

            if modname in sys.modules:
                break

        if last_j < len(name_parts):
            parent = None
            obj = sys.modules[modname]
            for obj_name in name_parts[last_j:]:
                parent = obj
                obj = getattr(obj, obj_name)
            return obj, parent, modname
        else:
            return sys.modules[modname], None, modname
    except (ValueError, ImportError, AttributeError, KeyError) as e:
        raise ImportError(*e.args)

# autosummary.import_by_name = my_import_by_name
autosummary._import_by_name = _import_by_name


def configure(globals_dict, settings_module_name=None):
    """
    Same as :func:`atelier.sphinxconf.configure` but with an
    additional positional argument `settings_module_name` (the name of
    a Django settings module).  If this argument is specified, call
    :meth:`lino.startup` with it.
    """
    if settings_module_name is not None:
        from lino import startup
        startup(settings_module_name)
        print("Django started with DJANGO_SETTINGS_MODULE={}.".format(settings_module_name))
        from django.conf import settings
        print(settings.SITE.welcome_text())
        # print(settings.SITE.diagnostic_report_rst())

        # os.environ['DJANGO_SETTINGS_MODULE'] = settings_module_name

        # # Trigger loading of Djangos model cache in order to avoid
        # # side effects that would occur when this happens later while
        # # importing one of the models modules.
        # from django.conf import settings
        # settings.SITE.startup()

        # globals_dict.update(
        #     template_bridge=str('lino.sphinxcontrib.DjangoTemplateBridge'))

    intersphinx_mapping = globals_dict.setdefault('intersphinx_mapping', dict())
    intersphinx_mapping['cg'] = ('https://community.lino-framework.org/', None)
    intersphinx_mapping['book'] = ('https://www.lino-framework.org/', None)
    return sphinxconf.configure(globals_dict)



# from sphinx.jinja2glue import BuiltinTemplateLoader


# class DjangoTemplateBridge(BuiltinTemplateLoader):

#     """The :meth:`configure` method installs this as `template_bridge
#     <http://sphinx-doc.org/config.html#confval-template_bridge>`_ for
#     Sphinx.  It causes a template variable ``settings`` to be added
#     the Sphinx template context. This cannot be done using
#     `html_context
#     <http://sphinx-doc.org/config.html#confval-html_context>`_ because
#     Django settings are not pickleable.

#     """

#     def render(self, template, context):
#         from django.conf import settings
#         context.update(settings=settings)
#         return super(DjangoTemplateBridge, self).render(template, context)

#     def render_string(self, source, context):
#         from django.conf import settings
#         context.update(settings=settings)
#         return super(DjangoTemplateBridge, self).render_string(source, context)



# -*- coding: UTF-8 -*-
# Copyright 2013-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from django.utils.translation import ugettext_lazy as _
from django.contrib.humanize.templatetags.humanize import naturaltime

from lino.api import dd, rt
from lino.modlib.users.mixins import My
from etgen.html import E, tostring
# from lxml import etree
import lxml
# from lino.utils.soup import truncate_comment
from lino.core.gfks import gfk2lookup
from .roles import CommentsReader, CommentsUser, CommentsStaff
from .choicelists import CommentEvents
from lino import mixins
from lino.core.constants import CHOICES_BLANK_FILTER_VALUE


class CommentTypes(dd.Table):
    required_roles = dd.login_required(CommentsStaff)
    model = 'comments.CommentType'
    column_names = "name *"
    order_by = ["name"]

    insert_layout = """
    name
    id
    """

    detail_layout = """
    id name
    comments.CommentsByType
    """


class Comments(dd.Table):
    required_roles = dd.login_required(CommentsUser)

    model = 'comments.Comment'
    params_layout = "start_date end_date observed_event user reply_to"

    insert_layout = dd.InsertLayout("""
    reply_to owner owner_type owner_id
    # comment_type
    body
    private
    """, window_size=(60, 13), hidden_elements="reply_to owner owner_type owner_id")

    detail_layout = """
    id user created modified private
    reply_to owner owner_type owner_id comment_type
    body
    CommentsByComment
    """

    # html_parser = etree.HTMLParser()

    # ~ column_names = "id date user type event_type subject * body_html"
    # ~ column_names = "id date user event_type type project subject * body"
    # ~ hide_columns = "body"
    # ~ hidden_columns = frozenset(['body'])
    # ~ order_by = ["id"]
    # ~ label = _("Notes")

    @classmethod
    def get_simple_parameters(cls):
        for p in super(Comments, cls).get_simple_parameters():
            yield p
        yield "reply_to"

    @classmethod
    def get_table_summary(cls, obj, ar):
        # print("20190926 get_table_summary", ar.request)
        sar = cls.request_from(ar, master_instance=obj, limit=cls.preview_limit)
        # print "20170208", sar.limit
        # chunks = []
        # for o in sar.sliced_data_iterator:
        #     chunks.append(E.p(*o.as_summary_row(ar)))
        # return E.div(*chunks)
        chunks = [o.as_summary_row(ar) for o in sar.sliced_data_iterator]
        html = '\n'.join(chunks)
        return "<div>{}</div>".format(html)

    @classmethod
    def get_comment_header(cls, comment, ar):
        ch = []
        if (comment.modified - comment.created).total_seconds() < 1:
            t = _("Created " + comment.created.strftime('%Y-%m-%d %H:%M'))
        else:
            t = _("Modified " + comment.modified.strftime('%Y-%m-%d %H:%M'))
        ch.append(ar.obj2html(
            comment, naturaltime(comment.created), title=t))
        ch += [" ", _("by"), " ",
               ar.obj2html(comment.user, str(comment.user))]

        sar = cls.insert_action.request_from(ar)
        # print(20170217, sar)
        sar.known_values = dict(
            reply_to=comment, **gfk2lookup(
                comment.__class__.owner, comment.owner))
        if ar.get_user().is_authenticated:
            btn = sar.ar2button(None, _(" Reply "), icon_name=None)
            # btn.set("style", "padding-left:10px")
            ch += [" [", btn, "]"]

        # ch.append(' ')
        # ch.append(
        #     E.a(u"⁜", onclick="toggle_visibility('comment-{}');".format(
        #         comment.id), title=str(_("Hide")), href="#")
        # )
        return tostring(ch)

    # @classmethod
    # def as_li(cls, self, ar):
    #     # chunks = [ar.parse_memo(self.short_preview)]
    #     chunks = [self.short_preview]
    #
    #     by = _("{0} by {1}").format(
    #         naturaltime(self.created), str(self.user))
    #
    #     if (self.modified - self.created).total_seconds() < 1:
    #         t = _("Created " + self.created.strftime('%Y-%m-%d %H:%M') )
    #     else:
    #         t = _("Modified " + self.modified.strftime('%Y-%m-%d %H:%M') )
    #
    #     chunks += [
    #         " (", tostring(ar.obj2html(self, by, title=t)), ")"
    #     ]
    #     # if self.more_text:
    #     #     chunks.append(" (...)")
    #
    #     if ar.get_user().authenticated:
    #         sar = cls.insert_action.request_from(ar)
    #         # print(20170217, sar)
    #         sar.known_values = dict(reply_to=self, owner=self.owner)
    #         if sar.get_permission():
    #             btn = sar.ar2button(
    #                 None, _("Reply"), icon_name=None)
    #             chunks.append(' '+tostring(btn))
    #
    #
    #     html = ''.join(chunks)
    #     return html
    #     # return "<li>" + html + "</li>"


class MyComments(My, Comments):
    required_roles = dd.login_required(CommentsUser)
    auto_fit_column_widths = True
    order_by = ["-modified"]
    column_names = "id modified short_preview owner workflow_buttons *"


class AllComments(Comments):
    required_roles = dd.login_required(CommentsStaff)
    order_by = ["-created"]


class CommentsByX(Comments):
    required_roles = dd.login_required(CommentsReader)
    order_by = ["-created"]
    # order_by = ["-modified"]
    display_mode = "summary"
    card_layout = dd.DetailLayout("""
    card_summary
    CommentsByComment
    """)

    @classmethod
    def get_card_title(self, ar, obj):
        """Overrides the default behaviour
        """
        return ar.actor.get_comment_header(obj, ar)


# class MyPendingComments(MyComments):
#     label = _("My pending comments")
#     welcome_message_when_count = 0

#     @classmethod
#     def param_defaults(cls, ar, **kw):
#         kw = super(MyPendingComments, cls).param_defaults(ar, **kw)
#         kw.update(show_published=dd.YesNo.no)
#         return kw

class RecentComments(CommentsByX):
    required_roles = dd.login_required(CommentsReader)
    # required_roles = set([CommentsReader])
    column_names = "short_preview modified user owner *"
    stay_in_grid = True
    # order_by = ["-modified"]
    label = _("Recent comments")
    preview_limit = 10
    # display_mode = "summary"


class CommentsByType(CommentsByX):
    master_key = 'comment_type'
    column_names = "body created user *"


class CommentsByRFC(CommentsByX):
    master_key = 'owner'
    column_names = "body created user *"
    stay_in_grid = True
    display_mode = "list"
    simple_slavegrid_header = True
    insert_layout = dd.InsertLayout("""
    reply_to
    # comment_type
    body
    private
    """, window_size=(60, 13), hidden_elements="reply_to")

    @classmethod
    def param_defaults(cls, ar, **kw):
        kw = super(CommentsByRFC, cls).param_defaults(ar, **kw)
        kw['reply_to'] = CHOICES_BLANK_FILTER_VALUE
        return kw

    @classmethod
    def get_main_card(self, ar):
        ticket_obj = ar.master_instance
        sar = self.request_from(ar, master_instance=ticket_obj)
        html=ticket_obj.get_rfc_description(ar)
        sar = self.insert_action.request_from(sar)
        if sar.get_permission():
            btn = sar.ar2button(None, _("Write comment"), icon_name=None)
            html += "<p>" + tostring(btn) + "</p>"

        if html:
            return dict(
                card_title="Description",
                main_card_body=html, # main_card_body is special keyword
                id="[main_card]" # needed for map key in react...
            )
        else:
            return None

    @classmethod
    def get_table_summary(self, obj, ar):
        sar = self.request_from(ar, master_instance=obj)
        html = obj.get_rfc_description(ar)
        sar = self.insert_action.request_from(sar)
        if sar.get_permission():
            btn = sar.ar2button(None, _("Write comment"), icon_name=None)
            html += "<p>" + tostring(btn) + "</p>"

        html += "<ul>"
        for c in sar:
            html += "<li>{}<div id=\"{}\">{}</div></li>".format(
                self.get_comment_header(c, sar),
                "comment-" + str(c.id),
                ar.parse_memo(c.body))

        html += "</ul>"
        return ar.html_text(html)


class CommentsByMentioned(CommentsByX):
    # show all comments that mention the master instance
    master = dd.Model
    label = _("Mentions")

    # label = _("Comments mentioning this")

    @classmethod
    def get_filter_kw(cls, ar, **kw):
        mi = ar.master_instance
        if mi is None:
            return None
        Mention = rt.models.comments.Mention
        mkw = gfk2lookup(Mention.owner, mi)
        mentions = Mention.objects.filter(**mkw).values_list('comment_id', flat=True)
        # mentions = [o.comment_id for o in Mention.objects.filter(**mkw)]
        # print(mkw, mentions)
        # return super(CommentsByMentioned, cls).get_filter_kw(ar, **kw)
        kw.update(id__in=mentions)
        return kw


class CommentsByComment(CommentsByX):
    master_key = 'reply_to'
    display_mode = "list"
    stay_in_grid = True
    borderless_list_mode = True
    title = _("Replies")
    simple_slavegrid_header = True

    paginator_template = "PrevPageLink NextPageLink"
    hide_if_empty = True


def comments_by_owner(obj):
    return CommentsByRFC.request(master_instance=obj)


class Mentions(dd.Table):
    required_roles = dd.login_required(CommentsStaff)
    model = "comments.Mention"
    column_names = "comment owner created *"
    detail_layout = """
    id comment owner created
    """


class MentionsByOwner(Mentions):
    master_key = "owner"



# Copyright 2016-2018 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

"""This plugins installs two build methods for generating
:doc:`printable documents </admin/printing>` using `weasyprint
<http://weasyprint.org/>`__.

See :doc:`/specs/weasyprint`.

"""

# trying to get rid of disturbing warnings in
# https://travis-ci.org/lino-framework/book/jobs/260560833
import warnings
warnings.filterwarnings(
    "ignore", 'There are known rendering problems')
warnings.filterwarnings(
    "ignore", '@font-face support needs Pango >= 1.38')

import imagesize

from lino.api import ad, _


class Plugin(ad.Plugin):
    "See :doc:`/dev/plugins`."
    verbose_name = _("WeasyPrint")

    header_height = 20
    """Height of header in mm. Set to `None` if you want no header."""

    footer_height = 20
    """Height of footer in mm. Set to `None` if you want no header."""

    top_right_width = None
    """Width of top-right.jpg in mm. If not given, Lino computes it based on height.
    """

    top_right_image = None
    """The first image file found in config named either top-right.jpg or top-right.png."""

    header_image = None
    """The first image file found in config named either header.jpg or header.png."""

    margin_left = 17
    margin = 10

    def on_site_startup(self, site):
        if self.header_height:
            for ext in ('jpg', 'png'):
                fn = site.confdirs.find_config_file("top-right."+ext, "weasyprint")
                if fn:
                    self.top_right_image = fn
                    w, h = imagesize.get(fn)
                    if self.top_right_width is None:
                        self.top_right_width = self.header_height * w / h
                fn = site.confdirs.find_config_file("header."+ext, "weasyprint")
                if fn:
                    self.header_image = fn

        super(Plugin, self).on_site_startup(site)



# -*- coding: UTF-8 -*-
# Copyright 2012-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

""".. management_command:: status

Write a diagnostic status report about this :term:`Lino site`.

A functional replacement for the :manage:`diag` command.

"""

from django.core.management.base import BaseCommand
from django.conf import settings
from lino.api import dd, rt


class Command(BaseCommand):
    help = __doc__

    def handle(self, *args, **options):
        settings.SITE.startup()
        ar = rt.login()
        # fn = rt.find_config_file("status.jinja.rst", "jinja")
        fn = "jinja/status.jinja.rst"
        context = ar.get_printable_context()
        print(dd.plugins.jinja.render_jinja(ar, fn, context))



# -*- coding: UTF-8 -*-
# Copyright 2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from django.conf import settings
from django import http
from django.views.generic import View
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import ensure_csrf_cookie, csrf_protect

from lino.core import auth
from lino.core.requests import BaseRequest

from django.shortcuts import redirect


class Element(View):

    publisher_model = None

    def get(self, request, pk=None):
        ar = BaseRequest(request=request, renderer=settings.SITE.kernel.default_renderer, permalink_uris=True)
        obj = self.publisher_model.objects.get(id=pk) if pk is not None else None
        return self.publisher_model.get_publisher_response(ar, obj)


class Index(View):
    """
    Render the main page.
    """
    @method_decorator(ensure_csrf_cookie)
    def get(self, request, *args, **kw):
        # raise Exception("20171122 {} {}".format(
        #     get_language(), settings.MIDDLEWARE_CLASSES))
        ar = BaseRequest(request=request, renderer=settings.SITE.kernel.default_renderer, permalink_uris=True)
        env = settings.SITE.plugins.jinja.renderer.jinja_env
        template = env.get_template("publisher/index.pub.html")
        context = ar.get_printable_context(obj=self)
        response = http.HttpResponse(
            template.render(**context),
            content_type='text/html;charset="utf-8"')
        return response

class Login(View):
    """
    Render the main page.
    """
    @method_decorator(ensure_csrf_cookie)
    def get(self, request, *args, **kw):
        # raise Exception("20171122 {} {}".format(
        #     get_language(), settings.MIDDLEWARE_CLASSES))
        ar = BaseRequest(request=request, renderer=settings.SITE.kernel.default_renderer, permalink_uris=True)
        env = settings.SITE.plugins.jinja.renderer.jinja_env
        template = env.get_template("publisher/login.html")
        context = ar.get_printable_context(obj=self)
        response = http.HttpResponse(
            template.render(**context),
            content_type='text/html;charset="utf-8"')
        return response

    @method_decorator(ensure_csrf_cookie)
    def post(self, request, *args, **kw):
        username = request.POST.get('username')
        password = request.POST.get('password')
        user = auth.authenticate(
            request, username=username, password=password)
        auth.login(request, user, backend=u'lino.core.auth.backends.ModelBackend')

        return redirect("/")



# Copyright 2011-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

"""Utilities for authentication. Adapted from `django.contrib.auth`.

"""

from django.conf import settings
from django.utils.crypto import constant_time_compare
from django.utils.module_loading import import_string

from lino.utils import SimpleSingleton

class AnonymousUser(SimpleSingleton):
    """A singleton class whose instance will be assigned to the
    :attr:`user` attribute of anonymous incoming requests, similar to
    Django's approach.

    See also :attr:`lino.core.site.Site.anonymous_user_type`.

    """
    # authenticated = False

    is_authenticated = False
    """This is always `False`.
    See also :attr:`lino.modlib.users.User.is_authenticated`.
    """

    is_active = False

    email = None
    username = 'anonymous'
    modified = None
    partner = None
    language = None
    readonly = True
    pk = None
    id = None
    time_zone =  None
    notify_myself = False
    user_type = None
    is_anonymous = True

    def __init__(self):
        settings.SITE.startup()
        from lino.modlib.users.choicelists import UserTypes
        self.user_type = UserTypes.get_by_name(self.username, None)

    def __str__(self):
        return self.username

    def get_typed_instance(self, model):
        # 20131022 AttributeError at /api/outbox/MyOutbox : 'AnonymousUser'
        # object has no attribute 'get_typed_instance'
        return self

    def get_username(self):
        return self.username

    def get_preferences(self):
        """Return the preferences of this user. The returned object is a
        :class:`lino.core.userprefs.UserPrefs` object.

        """
        from lino.core import userprefs
        return userprefs.reg.get(self)

    def has_perm(self, perm, obj=None):
        return False

    def has_perms(self, perm_list, obj=None):
        for perm in perm_list:
            if not self.has_perm(perm, obj):
                return False
        return True

    def get_choices_text(self, request, actor, field):
        return str(self)



# Copyright 2009-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

"""Extends the possibilities for defining choices for fields of a
Django model.

- Context-sensitive choices (see :ref:`lino.dev.combo`)
- Non-limiting choices (`force_selection` False) :
  specify a pick list of suggestions but leave the possibility
  to store manually entered values
- :ref:`learning_combos`

Example values:

>>> import json
>>> s = '<a href="javascript:Lino.pcsw.Clients.detail.run(\
null,{ &quot;record_id&quot;: 116 })">BASTIAENSEN Laurent (116)</a>'
>>> print(json.dumps(GFK_HACK.match(s).groups()))
["pcsw.Clients", "116"]

>>> s = '<a href="javascript:Lino.cal.Guests.detail.run(\
null,{ &quot;record_id&quot;: 6 })">Gast #6 ("Termin #51")</a>'
>>> print(json.dumps(GFK_HACK.match(s).groups()))
["cal.Guests", "6"]


"""

import decimal
import datetime
from dateutil import parser as dateparser

import re
GFK_HACK = re.compile(r'^<a href="javascript:Lino\.(\w+\.\w+)\.detail\.run\(.*,\{ &quot;record_id&quot;: (\w+) \}\)">.*</a>$')

from django.db import models
from django.conf import settings

# from lino.api import rt
from lino.core import constants
from lino.core import fields
from lino.core.utils import getrqdata




# class DataError(Exception):
#     pass


class Converter(object):

    def __init__(self, field):
        self.field = field

    def convert(self, **kw):
        return kw


class LookupConverter(Converter):

    """
    A Converter for ForeignKey and ManyToManyField.
    If the lookup_field is a BabelField, then it tries all available languages.
    """

    def __init__(self, field, lookup_field):
        Converter.__init__(self, field)
        model = field.remote_field.model
        if lookup_field == 'pk':
            self.lookup_field = model._meta.pk
        else:
            self.lookup_field = model._meta.get_field(lookup_field)
        # self.lookup_field = lookup_field

    def lookup(self, value, **kw):
        model = self.field.remote_field.model
        if isinstance(value, model):
            return value
        return model.lookup_or_create(self.lookup_field, value, **kw)

        # if isinstance(self.lookup_field,babel.BabelCharField):
            # flt  = babel.lookup_filter(self.lookup_field.name,value,**kw)
        # else:
            # kw[self.lookup_field.name] = value
            # flt = models.Q(**kw)
        # try:
            # return model.objects.get(flt)
        # except MultipleObjectsReturned,e:
            # raise model.MultipleObjectsReturned("%s.objects lookup(%r) : %s" % (model.__name__,value,e))
        # except model.DoesNotExist,e:
            # raise model.DoesNotExist("%s.objects lookup(%r) : %s" % (model.__name__,value,e))


class DateConverter(Converter):

    def convert(self, **kw):
        value = kw.get(self.field.name)
        if not isinstance(value, datetime.date):
            if value:  # keep out empty strings
                if type(value) == int:
                    value = str(value)
                d = dateparser.parse(value)
                d = datetime.date(d.year, d.month, d.day)
                kw[self.field.name] = d
        return kw


class ChoiceConverter(Converter):

    """Converter for :class:`ChoiceListField
    <lino.core.choicelists.ChoiceListField>`.

    If you specify a string, then it can be a *value* or a *name*.

    """

    def convert(self, **kw):
        value = kw.get(self.field.name)

        if value is not None:
            if not isinstance(value, self.field.choicelist.item_class):
                # kw[self.field.name] = self.field.choicelist.get_by_value(value)
                kw[self.field.name] = self.field.choicelist.to_python(value)
                # if self.field.name == "vat_class":
                #     print("20191210 convert {} from {} --> {}".format(
                #         value, self.field.choicelist.items_dict, kw[self.field.name]))
        return kw


class DecimalConverter(Converter):

    def convert(self, **kw):
        value = kw.get(self.field.name)
        if value is not None:
            if not isinstance(value, decimal.Decimal):
                kw[self.field.name] = decimal.Decimal(value)
        return kw


class ForeignKeyConverter(LookupConverter):

    """Converter for ForeignKey fields."""

    def convert(self, **kw):
        value = kw.get(self.field.name)
        if value is not None:
            if value == '':
                value = None
            else:
                value = self.lookup(value)
            kw[self.field.name] = value
            # logger.info("20111213 %s %s -> %r", self.field.name,self.__class__,value)
        return kw


class GenericForeignKeyConverter(Converter):

    """Converter for GenericForeignKey fields."""

    def convert(self, **kw):
        value = kw.get(self.field.name)
        if value is not None:
            if value == '':
                value = None
            else:
                mo = GFK_HACK.match(value)
                if mo is not None:
                    actor = settings.SITE.models.resolve(mo.group(1))
                    pk = mo.group(2)
                    value = actor.get_row_by_pk(None, pk)
                    # ct = ContentType.objects.get_for_model(actor.model)
                    # value = self.lookup(value)
                else:
                    raise Exception("Could not parse %r" % value)
            kw[self.field.name] = value
        return kw


class ManyToManyConverter(LookupConverter):

    """Converter for ManyToMany fields."""
    splitsep = None

    # def lookup(self,value):
        # model = self.field.remote_field.model
        # try:
            # return model.objects.get(
              # **{self.lookup_field: value})
        # except model.DoesNotExist,e:
            # raise DataError("%s.objects.get(%r) : %s" % (
              # model.__name__,value,e))

    def convert(self, **kw):
        values = kw.get(self.field.name)
        if values is not None:
            del kw[self.field.name]
            l = [self.lookup(value)
                 for value in values.split(self.splitsep)]
            kw['_m2m'][self.field.name] = l
        return kw


def make_converter(f, lookup_fields={}):
    from lino.core.gfks import GenericForeignKey

    if isinstance(f, models.ForeignKey):
        return ForeignKeyConverter(f, lookup_fields.get(f.name, "pk"))
    if isinstance(f, GenericForeignKey):
        return GenericForeignKeyConverter(f)
    # if isinstance(f,fields.LinkedForeignKey):
        # return LinkedForeignKeyConverter(f,lookup_fields.get(f.name,"pk"))
    if isinstance(f, models.ManyToManyField):
        return ManyToManyConverter(f, lookup_fields.get(f.name, "pk"))
    if isinstance(f, models.DateField):
        return DateConverter(f)
    if isinstance(f, models.DecimalField):
        return DecimalConverter(f)
    from lino.core import choicelists
    if isinstance(f, choicelists.ChoiceListField):
        # if f.name == 'p_book':
            # print "20131012 b", f
        return ChoiceConverter(f)




class BaseChooser(object):
    pass


class FieldChooser(BaseChooser):

    def __init__(self, field):
        self.field = field


class ChoicesChooser(FieldChooser):

    def __init__(self, field):
        FieldChooser.__init__(self, field)
        self.simple_values = type(field.choices[0])


class Chooser(FieldChooser):
    """Holds information about the possible choices of a field.

    """
    #~ stored_name = None
    simple_values = False
    instance_values = True
    force_selection = True
    choice_display_method = None  # not yet used.
    can_create_choice = False

    def __init__(self, model, field, meth):
        FieldChooser.__init__(self, field)
        self.model = model
        #~ self.field = model._meta.get_field(fldname)
        self.meth = meth
        from lino.core.gfks import is_foreignkey
        from lino.core.choicelists import ChoiceListField
        if isinstance(field, ChoiceListField):
            self.simple_values = getattr(meth, 'simple_values', False)
            self.instance_values = getattr(meth, 'instance_values', True)
            self.force_selection = getattr(
                meth, 'force_selection', self.force_selection)
        elif is_foreignkey(field):
            pass
        elif isinstance(field, fields.VirtualField) and isinstance(field.return_type, models.ForeignKey):
            pass
        else:
            self.simple_values = getattr(meth, 'simple_values', False)
            self.instance_values = getattr(meth, 'instance_values', False)
            self.force_selection = getattr(
                meth, 'force_selection', self.force_selection)
        #~ self.context_params = meth.func_code.co_varnames[1:meth.func_code.co_argcount]
        self.context_params = meth.context_params
        #~ self.multiple = meth.multiple
        #~ self.context_params = meth.func_code.co_varnames[:meth.func_code.co_argcount]
        #~ print '20100724', meth, self.context_params
        #~ logger.warning("20100527 %s %s",self.context_params,meth)
        self.context_values = []
        self.context_fields = []
        for name in self.context_params:
            if name == "ar":
                continue
            f = self.get_data_elem(name)

            if f is None:
                raise Exception(
                    "No data element '%s' in %s "
                    "(method %s_choices)" % (
                        name, self.model, field.name))
            #~ if name == 'p_book':
                #~ print 20131012, f
            self.context_fields.append(f)
            self.context_values.append(name + "Hidden")
            #~ if isinstance(f,models.ForeignKey):
                #~ self.context_values.append(name+"Hidden")
            #~ else:
                #~ self.context_values.append(name)
        self.converters = []
        #~ try:
        for f in self.context_fields:
            cv = make_converter(f)
            if cv is not None:
                self.converters.append(cv)
        #~ except models.FieldDoesNotExist,e:
            #~ print e

        if hasattr(model, "create_%s_choice" % field.name):
            self.can_create_choice = True

        m = getattr(model, "%s_choice_display" % field.name, None)
        if m is not None:
            self.choice_display_method = m

    def __str__(self):
        return "Chooser(%s.%s,%s)" % (
            self.model.__name__, self.field.name,
            self.context_params)

    def create_choice(self, obj, text):
        m = getattr(obj, "create_%s_choice" % self.field.name)
        return m(text)

    def get_data_elem(self, name):
        """Calls :meth:`dd.Actor.get_data_elem` or
        :meth:`dd.Model.get_data_elem` or
        :meth:`dd.Action.get_data_elem`.

        """
        de = self.model.get_data_elem(name)
        if de is None:
            return self.model.get_param_elem(name)
        return de

    def __call__(self, *args, **kw):
        for i, v in enumerate(args):
            kw[self.context_fields[i]] = v
        return self.get_choices(**kw)

    def get_choices(self, **context):
        """Return a list of choices for this chooser, using keyword parameters
        as context.

        """
        args = []
        for varname in self.context_params:
            args.append(context.get(varname, None))
        return self.meth(*args)

    def get_request_choices(self, ar, tbl):
        """
        Return a list of choices for this chooser,
        using a HttpRequest to build the context.
        """
        kw = {
            "ar": ar,
        }
        # kw = {}

        # ba = tbl.get_url_action(tbl.default_elem_action_name)
        # 20120202
        if tbl.master_field is not None:
            from django.contrib.contenttypes.models import ContentType
            rqdata = getrqdata(ar.request)
            if tbl.master is not None:
                master = tbl.master
            else:
                mt = rqdata.get(constants.URL_PARAM_MASTER_TYPE)
                # ContentType = rt.models.contenttypes.ContentType
                try:
                    master = ContentType.objects.get(pk=mt).model_class()
                except ContentType.DoesNotExist:
                    master = None

            pk = rqdata.get(constants.URL_PARAM_MASTER_PK, None)
            if pk and master:
                try:
                    kw[tbl.master_field.name] = master.objects.get(pk=pk)
                except ValueError:
                    raise Exception(
                        "Invalid primary key %r for %s", pk, master.__name__)
                except master.DoesNotExist:
                    raise Exception("There's no %s with primary key %r" %
                                    (master.__name__, pk))

        for k, v in list(ar.request.GET.items()):
            kw[str(k)] = v

        # logger.info(
        #     "20130513 get_request_choices(%r) -> %r",
        #     tbl, kw)

        for cv in self.converters:
            kw = cv.convert(**kw)

        if tbl.known_values:
            kw.update(tbl.known_values)

        if False:  # removed 20120815 #1114
            #~ ar = tbl.request(ui,request,tbl.default_action)
            if ar.create_kw:
                kw.update(ar.create_kw)
            if ar.known_values:
                kw.update(ar.known_values)
            if tbl.master_key:
                kw[tbl.master_key] = ar.master_instance
            #~ if tbl.known_values:
                #~ kw.update(tbl.known_values)
        return self.get_choices(**kw)  # 20120918b

    def get_text_for_value(self, value, obj):
        m = getattr(self.field, 'get_text_for_value', None)
        if m is not None:  # e.g. lino.utils.choicelist.ChoiceListField
            return m(value)
        #~ raise NotImplementedError
        #~ assert not self.simple_values
        m = getattr(obj, "get_" + self.field.name + "_display", str)
        #~ if m is None:
            #~ raise Exception("")
        return m(value)
        #~ raise NotImplementedError("%s : Cannot get text for value %r" % (self.meth,value))


def uses_simple_values(holder, fld):
    "used by :class:`lino.core.store`"
    from lino.core.gfks import is_foreignkey
    if is_foreignkey(fld):
        return False
    # if isinstance(fld, models.OneToOneRel):
    #     return False
    if holder is not None and fld.name is not None:
        ch = holder.get_chooser_for_field(fld.name)
        if ch is not None:
            return ch.simple_values
    if fld.choices is None:
        return True
    choices = list(fld.choices)
    if len(choices) == 0:
        return True
    if type(choices[0]) in (list, tuple):
        return False
    return True


def _chooser(make, **options):
    #~ options.setdefault('quick_insert_field',None)
    def chooser_decorator(fn):
        def wrapped(*args):
            return fn(*args)
        cp = options.pop(
            'context_params',
            fn.__code__.co_varnames[1:fn.__code__.co_argcount])
        wrapped.context_params = cp
        for k, v in options.items():
            setattr(wrapped, k, v)
        return make(wrapped)
        # return classmethod(wrapped)
        # A chooser on an action must not turn it into a classmethod
    return chooser_decorator


def chooser(**options):
    "Decorator which turns the method into a chooser."
    return _chooser(classmethod, **options)


def noop(x):
    return x


def action_chooser(**options):
    return _chooser(noop, **options)

def get_choosers_dict(holder):
    d = holder.__dict__.get('_choosers_dict', None)
    if d is None:
        d = dict()
        setattr(holder, '_choosers_dict', d)
    return d

def check_for_chooser(holder, field):
    # holder is either a Model, an Actor or an Action.
    if isinstance(field, fields.DummyField):
        return
    d = get_choosers_dict(holder)
    ch = d.get(field.name, None)
    if ch is not None:
        # if ch.model is not holder:
        #     raise Exception("20200425 {} is not {}".format(holder, ch.model))
        return ch

    methname = field.name + "_choices"
    m = getattr(holder, methname, None)
    if m is not None:
        # if field.name.endswith('municipality'):
        #     print("20200524 check_for_chooser() found {} on {}", field.name, holder)
        # 20200425 fix theoretical bug
        # if field.name
        # if ch in d:
            # raise Exception(
            #     "Duplicate of chooser for {} in {}".format(field, holder))
        ch = Chooser(holder, field, m)
        d[field.name] = ch
        return ch
    # if field.name == 'city':
    #     logger.info("20140822 chooser for %s.%s", holder, field.name)



# Copyright 2010-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)
"""Utility functions used by :mod:`lino.modlib.extjs.views`.

"""

from django.conf import settings
from django.db import models
from django import http
from django.core import exceptions

from lino.core import constants
from lino.core import actors
from lino.utils.jsgen import py2js


def json_response_kw(**kw):
    return json_response(kw)


def json_response(x, content_type='application/json', status=200):
    if True:
        s = py2js(x)
    else:
        try:
            s = py2js(x)
        except Exception as e:
            raise Exception("Failed to render {!r} : {}".format(x, e))

    # Theroretically we should send content_type='application/json'
    # (http://stackoverflow.com/questions/477816/the-right-json-content-type),
    # but "File uploads are not performed using Ajax submission, that
    # is they are not performed using XMLHttpRequests. (...)  If the
    # server is using JSON to send the return object, then the
    # Content-Type header must be set to "text/html" in order to tell
    # the browser to insert the text unchanged into the document
    # body."
    # (http://docs.sencha.com/ext-js/3-4/#!/api/Ext.form.BasicForm)
    # See 20120209.

    return http.HttpResponse(s, content_type=content_type, status=status)
    #~ return HttpResponse(s, content_type='text/html')
    #~ return HttpResponse(s, content_type='application/json')
    #~ return HttpResponse(s, content_type='text/json')


def requested_actor(app_label, actor):
    """Return the requested actor, either directly or (if specified name
    is a model) that model's default table.

    """
    x = settings.SITE.models.get(app_label)
    # x = getattr(settings.SITE.models, app_label, None)
    if x is None:
        raise http.Http404("There's no app_label %r here" % app_label)
        # raise Exception("There's no app_label %r here" % app_label)
    cl = getattr(x, actor, None)
    if not isinstance(cl, type):
        raise http.Http404("%s.%s is not a class" % (app_label, actor))
        # raise http.Http404("%s.%s is not a class (but %r)" % (
        #     app_label, actor, cl))
    if issubclass(cl, models.Model):
        return cl.get_default_table()
    if not issubclass(cl, actors.Actor):
        #~ raise http.Http404("%r is not an actor" % cl)
        raise http.Http404("%r is not an actor" % cl)
    return cl


def action_request(app_label, actor, request, rqdata, is_list, **kw):
    # print(20160329, rqdata.keys())
    rpt = requested_actor(app_label, actor)
    action_name = rqdata.get(constants.URL_PARAM_ACTION_NAME, None)
    if not action_name:
        if is_list:
            action_name = rpt.default_list_action_name
        else:
            action_name = rpt.default_elem_action_name
    a = rpt.get_url_action(action_name)
    if a is None:
        raise http.Http404(
            "%s has no url action %r (possible values are %s)" % (
                rpt, action_name, rpt.get_url_action_names()))
    user = request.subst_user or request.user
    if True:  # False:  # 20130829
        if not a.get_view_permission(user.user_type):
            # print("20200110 {} {}".format(a.required, user.user_type.role.__class__))
            raise exceptions.PermissionDenied(
                "As %s you have no view permission for this action."
                % user.user_type)
                # The text of an Exception may not be
                # internationalized because some error handling code
                # may want to write it to a plain ascii stream.
    ar = rpt.request(request=request, action=a, rqdata=rqdata, **kw)
    return ar
